@@ -1,4 +1,4 @@
-# Copyright 2016-2021 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
+# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
 # ReFrame Project Developers. See the top-level LICENSE file for details.
 #
 # SPDX-License-Identifier: BSD-3-Clause
@@ -54,7 +54,7 @@ class TestCase:
         return iter([self._check, self._partition, self._environ])
 
     def __hash__(self):
-        return (hash(self.check.name) ^
+        return (hash(self.check.unique_name) ^
                 hash(self.partition.fullname) ^
                 hash(self.environ.name))
 
@@ -62,12 +62,14 @@ class TestCase:
         if not isinstance(other, type(self)):
             return NotImplemented
 
-        return (self.check.name == other.check.name and
+        return (self.check.unique_name == other.check.unique_name and
                 self.environ.name == other.environ.name and
                 self.partition.fullname == other.partition.fullname)
 
     def __repr__(self):
-        c, p, e = self.check.name, self.partition.fullname, self.environ.name
+        c = self.check.unique_name if self.check else None
+        p = self.partition.fullname  if self.partition else None
+        e = self.environ.name if self.environ else None
         return f'({c!r}, {p!r}, {e!r})'
 
     @property
@@ -215,6 +217,26 @@ class RegressionTask:
         return (self._failed_stage is not None and
                 not self._aborted and not self._skipped)
 
+    @property
+    def state(self):
+        if self.failed:
+            return 'fail'
+
+        if self.skipped:
+            return 'skip'
+
+        states = {
+            'startup': 'startup',
+            'setup': 'ready_compile',
+            'compile': 'compiling',
+            'compile_wait': 'ready_run',
+            'run': 'running',
+            'run_wait': 'completing',
+            'finalize': 'retired',
+            'cleanup': 'completed',
+        }
+        return states[self._current_stage]
+
     @property
     def failed_stage(self):
         return self._failed_stage
@@ -248,7 +270,9 @@ class RegressionTask:
             # we don't want to masquerade the self argument of our containing
             # function
             def __enter__(this):
-                if fn.__name__ != 'poll':
+                if fn.__name__ not in ('poll',
+                                       'run_complete',
+                                       'compile_complete'):
                     stage = self._current_stage
                     self._timestamps[f'{stage}_start'] = time.time()
 
@@ -257,7 +281,7 @@ class RegressionTask:
                 self._timestamps[f'{stage}_finish'] = time.time()
                 self._timestamps['pipeline_end'] = time.time()
 
-        if fn.__name__ != 'poll':
+        if fn.__name__ not in ('poll', 'run_complete', 'compile_complete'):
             self._current_stage = fn.__name__
 
         try:
@@ -284,6 +308,7 @@ class RegressionTask:
 
     def compile(self):
         self._safe_call(self.check.compile)
+        self._notify_listeners('on_task_compile')
 
     def compile_wait(self):
         self._safe_call(self.check.compile_wait)
@@ -300,6 +325,13 @@ class RegressionTask:
 
         return done
 
+    def compile_complete(self):
+        done = self._safe_call(self.check.compile_complete)
+        if done:
+            self._notify_listeners('on_task_compile_exit')
+
+        return done
+
     def run_wait(self):
         self._safe_call(self.check.run_wait)
         self.zombie = False
@@ -357,6 +389,13 @@ class RegressionTask:
         else:
             self.fail((type(exc), exc, None))
 
+    def info(self):
+        '''Return an info string about this task.'''
+        name = self.check.display_name
+        part = self.testcase.partition.fullname
+        env  = self.testcase.environ.name
+        return f'{name} @{part}+{env}'
+
 
 class TaskEventListener(abc.ABC):
     @abc.abstractmethod
@@ -367,10 +406,19 @@ class TaskEventListener(abc.ABC):
     def on_task_run(self, task):
         '''Called whenever the run() method of a RegressionTask is called.'''
 
+    @abc.abstractmethod
+    def on_task_compile(self, task):
+        '''Called whenever the compile() method of a RegressionTask is
+        called.'''
+
     @abc.abstractmethod
     def on_task_exit(self, task):
         '''Called whenever a RegressionTask finishes.'''
 
+    @abc.abstractmethod
+    def on_task_compile_exit(self, task):
+        '''Called whenever a RegressionTask compilation phase finishes.'''
+
     @abc.abstractmethod
     def on_task_skip(self, task):
         '''Called whenever a RegressionTask is skipped.'''
@@ -420,7 +468,7 @@ class Runner:
         return self._stats
 
     def runall(self, testcases, restored_cases=None):
-        num_checks = len({tc.check.name for tc in testcases})
+        num_checks = len({tc.check.unique_name for tc in testcases})
         self._printer.separator('short double line',
                                 'Running %d check(s)' % num_checks)
         self._printer.timestamp('Started on', 'short double line')
@@ -457,7 +505,7 @@ class Runner:
         rt = runtime.runtime()
         failures = self._stats.failed()
         while (failures and rt.current_run < self._max_retries):
-            num_failed_checks = len({tc.check.name for tc in failures})
+            num_failed_checks = len({tc.check.unique_name for tc in failures})
             rt.next_run()
 
             self._printer.separator(
@@ -477,29 +525,19 @@ class Runner:
         def print_separator(check, prefix):
             self._printer.separator(
                 'short single line',
-                '%s %s (%s)' % (prefix, check.name, check.descr)
+                '%s %s (%s)' % (prefix, check.unique_name, check.descr)
             )
 
+        self._printer.separator('short single line',
+                                'start processing checks')
         self._policy.enter()
         self._printer.reset_progress(len(testcases))
-        last_check = None
         for t in testcases:
-            if last_check is None or last_check.name != t.check.name:
-                if last_check is not None:
-                    print_separator(last_check, 'finished processing')
-                    self._printer.info('')
-
-                print_separator(t.check, 'started processing')
-                last_check = t.check
-
             self._policy.runcase(t)
 
-        # Close the last visual box
-        if last_check is not None:
-            print_separator(last_check, 'finished processing')
-            self._printer.info('')
-
         self._policy.exit()
+        self._printer.separator('short single line',
+                                'all spawned checks have finished\n')
 
 
 class ExecutionPolicy(abc.ABC):
